//
//GowinSynthesis Coding Template Of Inference 
//
//GWFPGA DSP

// =========Oooo=======================================oooO=======
//1.various width
//The below case will be inferred to DSP，
//a_width and b_width can be changed
// =========Oooo=======================================oooO=======

//Template1.1:
input [a_width-1:0]a;
input [b_width-1:0]b;
assign c=a*b;
			
// =========Oooo=======================================oooO=======
//2.signed/unsigned flag
//DSP has parameter ASIGN and BSIGN, which is the flag to 
//indicate the multiplication is signed or unsigned. 
// =========Oooo=======================================oooO=======

//Template2.1:
//ASIGN = 0, BSIGN = 0;
input [a_width-1:0]a;
input [b_width-1:0]b;
assign c=a*b;

//Template2.2:	
//ASIGN = 1, BSIGN = 0;
input signed [a_width-1:0]a;
input [b_width-1:0]b;
assign c=a*b;

//Template2.3:
//ASIGN = 0, BSIGN = 1;
input [a_width-1:0]a;
input signed [b_width-1:0]b;
assign c=a*b;

//Template2.4:
//ASIGN = 1, BSIGN = 1;
input signed [a_width-1:0]a;
input signed [b_width-1:0]b;
assign c=a*b;
		
// =========Oooo=======================================oooO=======
//3.register
//There are three different types of registers in DSP, 
//like input registers, pipeline registers and output register.
//These three types of register are represented in the following case:
// =========Oooo=======================================oooO=======

//Template3.1:
//In the above case, ina and inb are input register, pp_reg is pipeline register, out_reg is output register.
input clock;
input reset;
input ce;
reg [a_width-1:0] ina;
reg [b_width-1:0] inb;
reg unsigned [c_width-1:0] pp_reg;
reg unsigned [c_width-1:0] out_reg;
wire unsigned [c_width-1:0] mult_out;
always @(posedge clock) begin
	if(reset)begin
		ina<=0;
		inb<=0;
	end else begin
		if(ce)begin
			ina<=a;
			inb<=b;
		end
	end
end
assign mult_out=ina*inb;
always @(posedge clock) begin
	if(reset)begin
		pp_reg<=0;
	end else begin
		if(ce)begin
			pp_reg<=mult_out;
		end
	end
end
always @(posedge clock) begin
	if(reset)begin
		out_reg<=0;
	end else begin
		if(ce)begin
			out_reg<=pp_reg;
		end
	end
end
assign c = out_reg;

// =========Oooo=======================================oooO=======
//4.reset mode
//The DSP parameter RESET_MODE can be set as sync or async, 
//which depends on the reg`s RESET_MODE absorbed by DSP.
//The details are as follows:
// =========Oooo=======================================oooO=======
//Template4.1:
//The input registers ina and inb is sync, then the DSP's RESET_MODE is sync;
input clock;
input reset;
input ce;
always @(posedge clock) begin
	if(reset)begin
		ina<=0;
		inb<=0;
	end else begin
		if(ce)begin
			ina<=a;
			inb<=b;
		end
	end
end
assign mult_out=ina*inb;
			
//Template4.2:
//The input registers ina and inb is async, then the DSP's RESET_MODE is async.
input clock;
input reset;
input ce;
always @(posedge clock or posedge reset) begin
	if(reset)begin
		ina<=0;
		inb<=0;
	end else begin
		if(ce)begin
			ina<=a;
			inb<=b;
		end
	end
end
assign mult_out=ina*inb;

// =========Oooo=======================================oooO=======
//5.adder chain after multiplications
//when two multiplication's result add together, 
//the two multiplications and adder will be inferred to one DSP
// =========Oooo=======================================oooO=======

//Template5.1:
module top(a0, a1, b0, b1, s);
parameter a0_width=18;
parameter a1_width=18;
parameter b0_width=18;
parameter b1_width=18;
parameter s_width=37;
input unsigned [a0_width-1:0] a0;
input unsigned [a1_width-1:0] a1;
input unsigned [b0_width-1:0] b0;
input unsigned [b1_width-1:0] b1;
output unsigned [s_width-1:0] s;
assign s=a0*b0+a1*b1;
endmodule

// =========Oooo=======================================oooO=======
//6.Accumulation
//DSP will be inferred for the below accumulate logic
// =========Oooo=======================================oooO=======

//Template6.1:
module acc(a, b, s, accload, reset, ce, clock);
parameter a_width=36; //18 36
parameter b_width=18; //18 36
parameter s_width=54; //54
input unsigned [a_width-1:0] a;
input unsigned [b_width-1:0] b;
input accload, reset, ce, clock;
output unsigned [s_width-1:0] s;
wire unsigned [s_width-1:0] s_sel;
wire unsigned [s_width-1:0] p;
reg [s_width-1:0] s;
assign p = a*b ;
assign s_sel = (accload == 1'b1) ? s : 54'h0000000;
always @(posedge clock)
begin
	if(reset)begin
		s <= 0;
	end else begin
		if(ce)begin
			s <= s_sel + p;
		end
	end
end
endmodule

// =========Oooo=======================================oooO=======
//7.Pre-adder
//Pre-adder is supported for one 9/18-bit operator. 
//Pre-adder’s output must feed on DSP.
//PADD and DSP would be inferred for the below case. 
// =========Oooo=======================================oooO=======

//Template7.1:
module top(a, bX, bY, p);
parameter a_width=36; //9 18 36
parameter b_width=18; //9 18
parameter p_width=54; //18 36 54
input [a_width-1:0] a;
input [b_width-1:0] bX, bY;
output [p_width-1:0] p;
wire [b_width-1:0] b;

assign b = bX + bY;
assign p = a*b;
endmodule

			
//GWFPGA RAM
// =========Oooo=======================================oooO=======
// 1. Register array width and depth could be recognized from
// following RTL coding styles, and used to match potential
// SRAM/ROM primitives. 
// =========Oooo=======================================oooO=======

// Register Array Width/Depth

//Template1.1: 
reg [wordsize-1:0] array_name [arraysize-1:0];
//Template1.2: 
reg [wordsize-1:0] array_1;
reg [wordsize-1:0] array_2;
... ...
reg [wordsize-1:0] array_n;

// =========Oooo=======================================oooO=======
//2.There are four port modes: Single Port (SP), Semi Dual Port 
//(SDP,rSDP), Dual Port (DP) and Read Only Memory (ROM,rROM). 
// =========Oooo=======================================oooO=======

//Template2.1: 
//Single Port mode can write to or read from one port at one clock edge. 
//As the result the following example can be inferred to SP BSRAM (2Kx8). 

//SP(BIT_WIDTH = 8) BSRAM
parameter addr_width = 11;
parameter data_width = 8;
reg [data_width - 1 : 0] mem [(2**addr_width)- 1:0];
always @(posedge clk)
begin
	if(wre)
		mem[addr] <= data_in;
	data_output <= mem[addr];
end

//Template2.2: 
//Semi-Dual Port mode has one read port and one write port, read and
//write address are different. It only supports one port read and another 
//port write. As the result the following example can be inferred to SDP 
//B-SRAM (1Kx16).
//SDP (BIT_WIDTH = 16) BSRAM
parameter addr_width = 10;
parameter data_width = 16;
reg [data_width - 1 : 0] mem [(2**addr_width)-1:0];
always @(posedge wclk)
	if(wre)
		mem[waddr] <= data_in;
always @(posedge rclk)
		data_output <= mem[raddr];

//Template2.3: 
//Dual Port mode has two separates ports, and each of them can do both 
//read and write. Each port has its own address port and clock. As the 
//result the following example can be inferred to DPX9 B-SRAM (2Kx9).

//DPX9 (BIT_WIDTH = 9) BSRAM
parameter addr_width = 11;
parameter data_width = 9;
reg [data_width - 1 : 0] mem [(2**addr_width)-1:0];
always @(posedge clka)
begin
	if(wrea)
		mem[addra] <= data_ina;
	data_outputa <= mem[addra]
end
always @(posedge clkb)
begin
	if(wreb)
		mem[addrb] <= data_inb;
	data_outputb <= mem[addrb];
end

//Template2.4: 
//ROM mode has only one read port. The example below can be inferred to 
//ROM16 S-SRAM. 

//ROM16
reg [15:0] mem/*synthesis syn_romstyle="distributed_rom"*/;
initial begin
		mem=$random;
end
always @(posedge clk)
		data_output <= mem[addr];

//Template2.5:
//An equivalent coding style, which combines case statement and 
//constant values, must be supported too. See the example below.

//ROM
reg [5:0] data_out;
always @ (posedge clock)
begin
case (address)
8'b00000000: data_out = 6'b101111;
8'b00000001: data_out = 6'b110110;
...
8'b11111110: data_out = 6'b000001;
8'b11111111: data_out = 6'b101010;
endcase
end

// =========Oooo=======================================oooO=======
//3. Write modes include 1) normal model, 2) write-through 
//mode and 3) read-before-write mode. Parameter WRITE_MODE can be set
// to 00, 01 or 10 respectively.
// =========Oooo=======================================oooO=======

//Template3.1:
//In normal mode, the output of a port won't change when the port is
// written. Here's the coding style for normal mode.
 
//Normal mode, WRITE_MODE = 2'b00
always@(posedge clk)
     	if(wre)
    		mem[addr] <= data_in;
		else
        	data_out <= mem[addr];

//Template3.2:
//Write-through mode allows the output data updated with the input data.
// The coding style is shown below. 
//Write-through mode, WRITE_MODE = 2'b01
always@(posedge clk)
   	 	if(wre)
		begin
    		mem[addr] <= data_in;
        	data_out <= data_in;
		end
    	else
        	data_out <= mem[addr];

//Template3.3:
//In read-before-write mode the output data gets the existing value 
//stored in the write address, and then the new value writes to the address.
// All the examples in port mode section are in read-before-write mode.
// Here is the coding style for read-before-write mode.

//Read-before-write mode, WRITE_MODE = 2'b10
always@(posedge clk)
begin
	if(wre)
		mem[addr] <= data_in;
	data_output <= mem[addr];
end

// =========Oooo=======================================oooO=======
//4. B-SRAM has extra level of pipeline registers available on data 
//output for each port. RAM inferencing should pack registers into 
//B-SRAM primitives if possible. When pipeline register is utilized, 
//READ_MODE is set to 1. An example is shown below.
// =========Oooo=======================================oooO=======

//Template4.1:
//READ_MODE = 1'b1
always@(posedge clk)
begin
	if(wre)
		mem[addr] <= data_in;
	data_output <= mem[addr];
	pipeline_output <= data_output;
end

//Template4.2:
//READ_MODE = 1'b0
always@(posedge clk)
begin
	if(wre)
		mem[addr] <= data_in;
	data_output <= mem[addr];
end

// =========Oooo=======================================oooO=======
//5. The revised BSRAM is a type of BSRAM either, it is the same with
// common BSRAM except write control signal. 
// =========Oooo=======================================oooO=======

//Template5.1:
//SDP
reg mem [1024:0];
reg dout_reg;
always @(posedge clka)
if(wre)
        mem[addra] <= data_in;
always@(posedge clkb) begin
	dout_reg <= mem[addrb];
	data_out <= dout_reg;
end

//Template5.2:
//rSDP.
reg mem [1024:0];
reg dout_reg;
always @(posedge clka)
        mem[addra] <= data_in;
always@(posedge clkb) begin
	dout_reg <= mem[addrb];
	data_out <= dout_reg;
end

// =========Oooo=======================================oooO=======
//6.RAM inferencing could support initial memory content in two ways
// =========Oooo=======================================oooO=======

//Template6.1:
//Initial values in RTL code
reg [7:0] mem [0:31];
integer i;
initial 
begin
for (i = 0; i < 32; i = i + 1)
mem[i] = i[7:0];
end

//Template6.2:
//initial values in a file,file should be specified by absolute path
reg [7:0] ram[0:15];
initial
begin
$readmemb("E:\project\src\ram.txt", ram);
End

// =========Oooo=======================================oooO=======
//7. Except the synchronous read coding style in the previous chapter,
// another way to describe the functionality is to put the register on 
//read address. RAM inferencing should support both when they are 
//equivalent. Below is an example of the alternative synchronous read 
//coding style.
// =========Oooo=======================================oooO=======

//Template7.1:
//ROM16 synchronous read
reg [15:0] mem/*synthesis syn_romstyle= "distributed_rom" */;
reg [3:0] addr_reg;
initial begin
		mem=$random;
end
always @(posedge clk)
		addr_reg <= addr;
	assign data_output=mem[addr_reg];

// =========Oooo=======================================oooO=======
//8.When the memory size in user design doesn't match B-SRAM/S-SRAM 
//natural depth or width, RAM inferencing need to properly tie-off unused
// data /address bits on RAM primitives and/or cascade primitives to 
//form bigger memory size. For example, 16x8 ROM functionality should 
//infer to eight ROM16 primitives fed by the same 4-bit address; 2Kx16 
//RAM should be implemented by two 1Kx16 primitives and a MUX 
//operator controlled by the address MSB. 
// =========Oooo=======================================oooO=======

//Template8.1:
//Below is an example for cascaded to 78 SP.
reg [38:0] mem [2**15-1:0] /* synthesis syn_ramstyle = "BSRAM" */;
reg[38:0] data_out;
always@(posedge clk)
    if(wre == 1)
        data_out <= data_in;
    else if (wre == 0)
        data_out <= mem[addr];
always @(posedge clk)
    if (wre) mem[addr] <= data_in;
// =========Oooo=======================================oooO=======
//9. Shift register is a commonly-used functionality that can be 
//implemented by RAM and a simple counter. Therefore, RAM inferencing
// need to recognize shift register functionality and infer RAM primitives 
//and counters. Below is an example of 1x16 shift register should be 
//inferred as one RAM16S1 and a 4-bit counter. 
// =========Oooo=======================================oooO=======

//Template9.1:
//shift_1x16
reg [15:0] sr;
always @ (posedge clk)
begin
if (shift == 1'b1)
begin
sr[15:1] <= sr[14:0];
sr[0] <= sr_in;
end
end
assign sr_out = sr[15];

// =========Oooo=======================================oooO=======
//10.Decoder with registers is another commonly-used functionality that 
//can be implemented by RAM. Therefore, Ram inferencing need to 
//recognize decoder functionality and infer RAM primitives. Below is
// an example of decoder+registers should be inferred as 16 RAM16S4.
// =========Oooo=======================================oooO=======

//Template10.1:
//decoder
reg [63 : 0] mem_0,mem_1,mem_2,mem_3,mem_4,mem_5,mem_6,mem_7,mem_8,
mem_9,mem_10,mem_11,mem_12,mem_13,mem_14,mem_15;

initial
begin
	mem_0  = INIT_0;
	mem_1  = INIT_1;
	mem_2  = INIT_2;
	mem_3  = INIT_3;
	mem_4  = INIT_4;
	mem_5  = INIT_5;
	mem_6  = INIT_6;
	mem_7  = INIT_7;
	mem_8  = INIT_8;
	mem_9  = INIT_9;
	mem_10 = INIT_10;
	mem_11 = INIT_11;
	mem_12 = INIT_12;
	mem_13 = INIT_13;
	mem_14 = INIT_14;
	mem_15 = INIT_15;
end

assign spo[0]  = mem_0[a];
assign spo[1]  = mem_1[a];
assign spo[2]  = mem_2[a];
assign spo[3]  = mem_3[a];
assign spo[4]  = mem_4[a];
assign spo[5]  = mem_5[a];
assign spo[6]  = mem_6[a];
assign spo[7]  = mem_7[a];
assign spo[8]  = mem_8[a];
assign spo[9]  = mem_9[a];
assign spo[10] = mem_10[a];
assign spo[11] = mem_11[a];
assign spo[12] = mem_12[a];
assign spo[13] = mem_13[a];
assign spo[14] = mem_14[a];
assign spo[15] = mem_15[a];

always @(posedge clk)
begin
	mem_0[a] <= d[0];
	mem_1[a] <= d[1];
	mem_2[a] <= d[2];
	mem_3[a] <= d[3];
	mem_4[a] <= d[4];
	mem_5[a] <= d[5];
	mem_6[a] <= d[6];
	mem_7[a] <= d[7];
	mem_8[a] <= d[8];
	mem_9[a] <= d[9];
	mem_10[a] <= d[10];
	mem_11[a] <= d[11];
	mem_12[a] <= d[12];
	mem_13[a] <= d[13];
	mem_14[a] <= d[14];
	mem_15[a] <= d[15];
end
			
			